'''
Created on 06.06.2015

@author: Oliver Scheuss
'''


# xml reader
import xml.dom.minidom

from parse_addresses import AddressBook, Address, AddressChargerType, AddressIcon, AddressRegion

# regular expression matching
import re


class POI_Collection:

    pwr_re_str = "(\d+)\s*kW";
    
    plug_re_strs = \
    [
        ("chademo", "CHAdeMO"),
        ("ccs", "CCS")
    ]
    
    entry_name_lambda = lambda element: element.getElementsByTagName("name")[0].childNodes[0].data;
    entry_desc_lambda = lambda element: element.getElementsByTagName("desc")[0].childNodes[0].data;
    entry_lat_lambda = lambda element: element.getAttribute("lat");
    entry_lon_lambda = lambda element: element.getAttribute("lon");
    
    collection_entries = \
    { 
        ('name',    entry_name_lambda),
        ('desc',    entry_desc_lambda),
        ('lat',     entry_lat_lambda),
        ('lon',     entry_lon_lambda)
    }
    
    table_entries = \
    [
        ('Name'         , 'name'),
        ('Power'        , 'pwr'),
        ('Description'  , 'desc'),
        ('Latitude'     , 'lat'),
        ('Longitude'    , 'lon'),
        ('Plugs'        , 'plugs')
    ]
    
    def __init__(self):
        # init empty collection
        self.collection = [];
        
        return;
        
    def import_gpx(self, filename):
        DOMTree = xml.dom.minidom.parse(filename);
        collection = DOMTree.documentElement;
        
        # Get all POIs in the collection
        poi_tree = collection.getElementsByTagName("wpt");
        
        poi_collection = [];
        
        for element in poi_tree:
            element_entry = {};
            for entry_type in self.collection_entries:
                key                 = entry_type[0];
                value_lambda        = entry_type[1];
                value               = value_lambda(element);
                element_entry[key]  = value;
                
            poi_collection.append(element_entry);
        
        # post process to extract additional information
        self.post_process_pois(poi_collection);
        
        # append new list to collection
        self.collection += poi_collection;
        
        return;
    
    def export_navteq(self, in_data):
        book = AddressBook();
        
        if in_data != None:
            book.parse(in_data);
        else:
            book.make_book();
            # set region to Europe
            book.region = AddressRegion.Europe;
        
        # remove all old addresses
        book.clear();
        
        
        for entry in self.collection:
            lon = entry['lon'];
            lat = entry['lat'];
            name = entry['name'];
            
            public = True;
            show_icon = True;
            
            # extract plug names from list of tuples of plug name and plug power
            plug_list = [i[0] for i in entry['plugs']];
            if 'chademo' in plug_list:
                charger_type = AddressChargerType.CHAdeMO;
                icon_id = AddressIcon.EVSE_Block;
            else:
                charger_type = AddressChargerType.L2
                icon_id = AddressIcon.EVSE_Round;
            
            address = Address(longitude=lon, latitude=lat, name=name, charger_type=charger_type,
                              public=public, show_icon=show_icon, icon_id=icon_id);
            address.set_position(lon, lat);
            
            book.append(address);
            
        return book.unparse();
    
    def get_poi_table(self):
        
        # init table header
        table_header = [];
        
        for entry in self.table_entries:
            table_header.append(entry[0]);
        
        
        table = [];
        
        for element in self.collection:
            element_entry = [];
            for entry_part in self.table_entries:
                key     = entry_part[1];
                value   = element[key]
                element_entry.append(value);
            table.append(element_entry);
        
        return (table_header, table);
    
    def post_process_pois(self, poi_collection):
        postproc_entires = \
        [
            ('pwr',     self.extract_pwr),
            ('plugs',   self.extract_plugs)
        ]
        
        for item in poi_collection:
            for pp_action in postproc_entires:
                key         = pp_action[0];
                pp_lambda   = pp_action[1];
                item[key]   = pp_lambda(item);
        return;
    
    def extract_pwr(self, element):
        # search power rating in name
        name_str = element['name'];
        match = re.search(self.pwr_re_str, name_str)
        
        if match and match.group(1): 
            return float(match.group(1));
        
        # pwr not found in name, fallback:
        # find biggest power rating in description
        desc_str = element['desc'];
        match = re.search(self.pwr_re_str, desc_str)
        if match:
            max_pwr = 0.0;
            for pwr_str in match.groups():
                pwr = float(pwr_str);
                if pwr > max_pwr:
                    max_pwr = pwr;
            return max_pwr;
        
        else:
            return '';
    
    def extract_plugs(self, element):
        plugs = [];
        
        desc_str = element['desc'];
        
        for plug_type in self.plug_re_strs:
            plug_name = plug_type[0];
            plug_str = plug_type[1];
            
            re1_str = plug_str + "[\s\w]+\((\d+)\s*kW\)";
            re2_str = plug_str;
            
            match1 = re.search(re1_str, desc_str);
            match2 = re.search(re2_str, desc_str);
            
            if match1 and match1.group(1):
                plug = (plug_name, float(match1.group(1)));
                plugs.append(plug)
            elif match2:
                plug = (plug_name, '');
                plugs.append(plug)
        
        
        return plugs;
    
